
#lang racket

(provide display-normal-list)

(define (mlist->list l)
  (cond
   ((mpair? l) (cons (mlist->list (mcar l)) (mlist->list (mcdr l))))
   ((pair? l) (cons (mlist->list (car l)) (mlist->list (cdr l))))
   (else l)))

(define (display-normal-list x)
  (display (mlist->list x)))
